@@ -1,17 +1,16 @@
 #
 # field.rb
 #
-#   Copyright (c) 1998-1999 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
+#   Copyright (c) 1998-2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
 #
 #   This program is free software.
 #   You can distribute/modify this program under the terms of
 #   the GNU Lesser General Public License version 2 or later.
 #
 
+require 'sync'
 require 'delegate'
-
 require 'amstd/to_s'
-
 require 'tmail/mailp'
 require 'tmail/encode'
 
@@ -67,9 +66,12 @@ module TMail
     'z'   =>   0 * 60
   }
 
-  TSPECIAL = %_()<>@,.;:\\"/[]?=_
-  CONTROL  = "\\\x00-\\\x20\\\x7f-\\\xff"
-  INSECURE = /[#{Regexp.quote TSPECIAL}#{CONTROL}]/o
+  TSPECIAL     = %_()<>@,.;:\\"/[]?=_
+  CONTROL      = "\\\x00-\\\x19\\\x7f-\\\xff"
+  INSECURE     = /[#{Regexp.quote TSPECIAL}#{CONTROL} ]/o
+  INSECURE_PH  = /[#{Regexp.quote TSPECIAL}#{CONTROL}]/o
+
+  ENCODED = /=\?[^\s?=]+\?[qb]\?[^\s?=]+\?=/i
 
   class << self
 
@@ -103,33 +105,44 @@ module TMail
 
   def quote( str )
     if INSECURE === str then
-      %("#{str}")
+      '"' + str.gsub('"', '\\\\"') << '"'
     else
       str
     end
   end
 
+  def quote_phrase( str )
+    if INSECURE_PH === str then
+      '"' + str.gsub('"', '\\\\"') << '"'
+    else
+      str
+    end
+  end
+
+  def encoded?( str )
+    ENCODED === str
+  end
+
 
   # internal use
 
-  def define_header_to_s( mod, eoptarg, doptarg )
+  def define_header_to_s( mod, use_limit )
     mod.module_eval %^
       def encoded( eol = "\r\n", charset = 'j',
-                   ret = '' #{eoptarg ? ',' + eoptarg : ''} )
-        ret = ''
-        v = ::TMail::HFencoder.new( ret, eol, charset, limit )
-        accept v
-        v.write_in
+                   ret = ''#{use_limit ? ', limit = 72' : ''} )
+        stra = ::TMail::HFencoder.new( ret, eol, charset,
+                 #{use_limit ? 'limit' : 'nil'}  )
+        accept stra
+        stra.write_in
         ret
       end
       alias to_s encoded
 
       def decoded( eol = "\n", charset = 'e',
-                   ret = '' #{doptarg ? ',' + doptarg : ''} )
-        ret = ''
-        v = ::TMail::HFdecoder.new( ret, charset )
-        accept v
-        v.write_in
+                   ret = ''#{use_limit ? ', limit = 72' : ''} )
+        stra = ::TMail::HFdecoder.new( ret, charset )
+        accept stra
+        stra.write_in
         ret
       end
       alias inspect decoded
@@ -147,15 +160,17 @@ class HeaderField
     @body = fbody
     @strict = strict
 
+    @illegal = false
     @parsed = false
     @parsing = false
-    @written = false
+    # @sync = Sync.new
   end
 
 
   R     = (1 << 0)  # readable
   W     = (1 << 1)  # writable
   NODUP = (1 << 2)  # dup?
+  NOEQL = (1 << 3)  # used by eql? ?
 
   class << self
 
@@ -177,17 +192,11 @@ class HeaderField
 
 
     def parse_on_rw
-      @parse_on_rw = true
+      @attrs = nil
     end
 
     def parse_on_create
-      @parse_on_rw = false
-      module_eval %^
-        def initialize( fname, fbody, strict )
-          super
-          parse
-        end
-      ^
+      @attrs = nil
     end
 
     def header_attr( name, type, flags, ali = nil )
@@ -196,57 +205,32 @@ class HeaderField
       r = (flags & R != 0)
       w = (flags & W != 0)
       c = (flags & NODUP != 0)
+      a = (flags & NOEQL != 0)
+
+      unless a then
+        if @attrs then
+          @attrs.push name
+        else
+          @attrs = [name]
+        end
+      end
 
       if r then
-        parse = if @parse_on_rw then
-                  %-
-                    unless @parsed or @parsing then
-                      parse
-                    end
-                  -
-                else
-                  ''
-                end
-        write = unless w then
-                  '@written = true'
-                else
-                  ''
-                end
-        retrn = if c then
-                  "@#{name}"
-                else
-                  "@#{name} and @#{name}.dup"
-                end
-        module_eval sprintf( %-
+        module_eval %-
           def #{name}
-            %s
-            %s
-            %s
+            ensure_parsed
+            @#{name}
           end
-        -, parse, write, retrn )
+        -
       end
 
       if w then
-        parse = if @parse_on_rw then
-                  %-
-                    unless @parsed or @parsing then
-                      parse
-                      @written = true
-                    end
-                  -
-                else
-                  %-
-                    unless @parsing then
-                      @written = true
-                    end
-                  -
-                end
-        module_eval sprintf( %-
+        module_eval %-
           def #{name}=( arg )
-            %s
+            ensure_parsed
             @#{name} = arg
           end
-        -, parse )
+        -
       end
 
       if ali then
@@ -257,57 +241,112 @@ class HeaderField
         end
       end
     end
+
+    def redefine_eql
+      bug! unless @attrs
+
+      eqls = @attrs.collect {|at| sprintf( '%s.eql?(other.%s)', at, at ) }
+      module_eval %-
+        def eql?( other )
+          super and
+          #{eqls.join(' and ')}
+        end
+      -
+
+      eqls = @attrs.collect {|at| sprintf( '%s == other.%s', at, at ) }
+      module_eval %-
+        def ==( other )
+          super and #{eqls.join(' and ')}
+        end
+      -
+    end
+
+    def empty_able
+      bug! unless @attrs
+
+      module_eval %-
+        def empty?
+          ensure_parsed
+          if @illegal then
+            /\A\s*\z/ === @body
+          else
+            not ( #{@attrs.join(' or ')} )
+          end
+        end
+      -
+    end
         
   end
 
-  # abstract parse
-
 
   def hash
     @name.hash
   end
 
   def eql?( oth )
-    self.type === oth and @name == oth.name
+    @name.downcase == oth.name.downcase
   end
 
-  def ==( oth )   # don't alias
-    eql? oth
+  alias == eql?
+
+
+  def illegal?
+    @illegal
   end
-    
+
+  def empty?
+    ensure_parsed
+    @illegal and (/\A\s*\z/ === @body)
+  end
+
 
   def name
     @name.dup
   end
 
-
   def body
-    ret = ''
-    v = HFdecoder.new( ret )
-    if @written then
-      do_accept v
-    else
-      v.header_body @body
-    end
+    ensure_parsed
+    v = HFdecoder.new( '' )
+    do_accept v
     v.write_in
-    ret
   end
 
+  def body=( str )
+    @body = str
+    @parsed = false
+    @illegal = false
+  end
 
-  def accept( visitor )
-    visitor.header_name @name
-    unless @written then
-      visitor.header_body @body
-    else
-      do_accept visitor
+  def accept( strategy )
+    ensure_parsed
+    strategy.header_name @name
+    do_accept strategy
+  end
+
+  ::TMail.define_header_to_s self, true
+
+
+  private
+
+  def ensure_parsed
+    begin
+      Thread.critical = true   #@sync.lock :EX
+      unless @parsed or @parsing then
+        @parsing = true
+        Thread.critical = false
+        parse
+        Thread.critical = true
+        @parsing = false
+        @parsed = true
+      end
+    ensure
+      Thread.critical = false   # @sync.unlock
     end
   end
 
-  ::TMail.define_header_to_s self, 'limit = 72', nil
+  # abstract parse
 
-  def do_accept( visitor )
-    visitor.header_body body
-  end
+  # abstract do_accept
 
 end
 
@@ -315,33 +354,36 @@ end
 
 # string type ---------------------------------------------
 
+module NoAttrHeader
+
+      def append_features( mod ) super; mod.
+  
+  header_attr :body, String, R | W    ; mod.
+  redefine_eql                          end
+
+  private
+
+  def do_accept( stra )
+    stra.text @body
+  end
+
+  def parse
+    init_real
+    TMail::HFdecoder.decode( @body )
+  end
+
+end
+
+
 class StringH < HeaderField
 
-  def initialize( fname, fbody, strict )
-    super
-    @decoded = nil
-    @written = true
-  end
+  parse_on_rw
 
-  def body
-    unless @decoded then
-      @decoded = HFdecoder.decode( @body )
-    end
-    @decoded.dup
-  end
+  include NoAttrHeader
 
-  def body=( str )
-    @body = nil
-    @decoded = str
-  end
+  private
 
-  def eql?( oth )
-    super and
-    body == oth.body
-  end
-
-  def do_accept( visitor )
-    visitor.text @body || @decoded
+  def init_real
   end
 
 end
@@ -353,29 +395,34 @@ class StructH < HeaderField
 
   parse_on_rw
 
-  header_attr :comments, Array, R | NODUP
+  header_attr :comments, Array, R | NODUP | NOEQL
 
 
   private
 
-  def init
+  def init_real
     @comments = []
+    init
   end
 
   def parse
-    @parsing = true
-    init
-    if @strict then
+    pre = nil
+
+    begin
+      init_real
       Mailp.parse( @body, self, nil )
-    else
-      begin
-        Mailp.parse( @body, self, nil )
-      rescue ParseError, ScanError
-        ;
+    rescue ParseError, ScanError
+      if not pre and ::TMail.encoded?( @body ) then
+        pre = @body
+        @body = HFdecoder.decode( pre )
+        retry
+      elsif pre then
+        @body = pre
       end
+
+      @illegal = true
+      raise if @strict
     end
-    @parsing = false
-    @parsed = true
   end
 
 end
@@ -385,36 +432,11 @@ end
 
 class UnknownH < StructH
 
-  def initialize( fname, fbody, strict )
-    super
-    @decoded = nil
-    @written = true
-  end
-
-  def body
-    unless @decoded then
-      @decoded = Bencode.decode( @body )
-    end
-    @decoded.dup
-  end
-
-  def body=( str )
-    @body = nil
-    @decoded = str
-  end
-
-  def eql?( oth )
-    super and
-    body == oth.body
-  end
+  include NoAttrHeader
 
   private
 
-  def parse
-  end
-
-  def do_accept( visitor )
-    visitor.text @body || @decoded
+  def init
   end
 
 end
@@ -430,16 +452,20 @@ class DateH < StructH
 
   def date=( arg )
     @date = arg.localtime
-    @written = true
   end
 
-  def eql?( oth )
-    super and
-    @date == oth.date
+  redefine_eql
+  empty_able
+
+
+  private
+
+  def init
+    @date = nil
   end
 
-  def do_accept( visitor )
-    visitor.meta ::TMail.time2str( @date )
+  def do_accept( strategy )
+    strategy.meta ::TMail.time2str( @date )
   end
 
 end
@@ -454,33 +480,34 @@ class RetpathH < StructH
   header_attr :route, Array,  R | NODUP, [:routes]
   header_attr :addr,  String, R | W
 
-  def eql?( oth )
-    super and
-    self.addr == oth.addr and
-    self.routes == oth.routes
-  end
+  redefine_eql
+  empty_able
 
 
   private
 
   def init
-    super
     @route = []
+    @addr = nil
   end
 
-  def do_accept( visitor )
-    visitor.meta '<'
-    unless @route.empty? then
-      last = @routes[-1]
+  def do_accept( stra )
+    stra.meta '<'
+    unless route.empty? then
+      first = true
       @route.each do |i|
         s = '@' + i
-        s << ',' unless i.equal? last
-        visitor.meta
+        if first then
+          first = false
+        else
+          s << ','
+        end
+        stra.meta
       end
-      visitor.meta ':'
+      stra.meta ':'
     end
-    visitor.meta @addr
-    visitor.meta '>'
+    stra.meta @addr
+    stra.meta '>'
   end
 
 end
@@ -504,9 +531,29 @@ class Address
   alias routes route
   attr_writer :route   # internal use only
 
+
+  def Address.join( arr )
+    arr.collect {|i| ::TMail.quote i }.join('.')
+  end
+
+  def local
+    @local.collect {|i| ::TMail.quote i }.join('.')
+  end
+
+  def domain
+    if @domain then
+      @domain.collect {|i| ::TMail.quote i }.join('.')
+    else
+      nil
+    end
+  end
+
   def address
-    s = @local.collect {|i| ::TMail.quote i }.join('.')
-    s << '@' << @domain.collect {|i| ::TMail.quote i }.join('.') if @domain
+    s = local
+    if d = domain then
+      s << '@' << d
+    end
+
     s
   end
 
@@ -538,23 +585,23 @@ class Address
     n
   end
 
-  def accept( visitor )
+  def accept( stra )
     spec_p = !@phrase and @route.empty?
 
     if @phrase then
-      visitor.text @phrase
-      visitor.spc
+      stra.phrase @phrase
+      stra.spc
     end
     tmp = spec_p ? '' : '<'
     unless @route.empty? then
-      tmp << @route.collect {|i| '@' + i }.join( ',' ) << ':'
+      tmp << @route.collect {|i| '@' + i }.join(',') << ':'
     end
     tmp << address
     tmp << '>' unless spec_p
-    visitor.meta tmp
+    stra.meta tmp
   end
 
-  ::TMail.define_header_to_s self, nil, nil
+  ::TMail.define_header_to_s self, false
 
 end
 
@@ -585,18 +632,24 @@ class AddressGroup < DelegateClass( Array )
     end
   end
 
-  def accept( visitor )
-    visitor.text @name
-    visitor.meta ':'
-    visitor.spc
-    last = self[-1]
+  def accept( stra )
+    stra.phrase @name
+    stra.meta ':'
+    stra.spc
+    first = true
     bare_each do |mbox|
-      mbox.accept visitor
-      visitor.meta mbox.equal?(last) ? ';' : ','
+      if first then
+        first = false
+      else
+        stra.meta ','
+      end
+      stra.spc
+      mbox.accept stra
     end
+    stra.meta ';'
   end
 
-  ::TMail.define_header_to_s self, nil, nil
+  ::TMail.define_header_to_s self, false
 
 end
 
@@ -609,14 +662,26 @@ class SaddrH < StructH
 
   header_attr :addr, Address, R | W
 
-  def eql?( oth )
-    super and
-    self.addr == oth.addr
+  redefine_eql
+  empty_able
+
+
+  private
+
+  def init
+    @addr = nil
   end
 
-  def do_accept( visitor )
-    @addr.accept visitor
-    visitor.comments @comments unless @comments.empty?
+  def do_accept( stra )
+    @addr.accept stra
+    unless @comments.empty? then
+      @comments.each do |c|
+        stra.spc
+        stra.meta '('
+        stra.text c
+        stra.meta ')'
+      end
+    end
   end
 
 end
@@ -633,37 +698,33 @@ class MaddrH < StructH
 
   header_attr :addrs, Array, R | NODUP
 
-  def eql?( oth )
-    super and
-    addrs == oth.addrs
-  end
+  redefine_eql
+  empty_able
 
 
   private
 
   def init
-    super
     @addrs = []
   end
 
-  def do_accept( visitor )
+  def do_accept( stra )
     first = true
-    last = @addrs[-1]
     @addrs.each do |a|
       if first then
         first = false
       else
-        visitor.spc
+        stra.meta ','
+        stra.spc
       end
-      a.accept visitor
-      visitor.meta ',' unless a.equal? last
+      a.accept stra
     end
     unless @comments.empty? then
-      visitor.spc
       @comments.each do |c|
-        visitor.meta '('
-        visitor.text c
-        visitor.meta ')'
+        stra.spc
+        stra.meta '('
+        stra.text c
+        stra.meta ')'
       end
     end
   end
@@ -673,7 +734,7 @@ end
 class MmboxH < MaddrH
 
 =begin
-  def do_accept( visitor )
+  def do_accept( stra )
     first = true
     last = @addrs[-1]
   end
@@ -690,10 +751,9 @@ class RefH < StructH
 
   header_attr :refs, Array, R | NODUP
 
-  def eql?( oth )
-    super and
-    self.refs == oth.refs
-  end
+  redefine_eql
+  empty_able
+
 
   def each_msgid
     refs.each do |i|
@@ -711,22 +771,21 @@ class RefH < StructH
   private
 
   def init
-    super
     @refs = []
   end
 
-  def do_accept( visitor )
+  def do_accept( stra )
     first = true
-    @refs.each do |i|
+    refs.each do |i|
       if first then
         first = false
       else
-        visitor.spc
+        stra.spc
       end
       if ::TMail.msgid? i then
-        visitor.meta i
+        stra.meta i
       else
-        visitor.text i
+        stra.phrase i
       end
     end
   end
@@ -742,26 +801,26 @@ class KeyH < StructH
 
   header_attr :keys, Array, R | NODUP
 
-  def eql?( oth )
-    super and
-    self.keys == oth.keys
-  end
+  redefine_eql
+  empty_able
 
 
   private
 
   def init
-    super
     @keys = []
   end
 
-  def do_accept( visitor )
-    save = @keys.pop
-    @keys.each do |i|
-      visitor.meta i + ','
+  def do_accept( stra )
+    first = true
+    keys.each do |i|
+      if first then
+        first = false
+      else
+        stra.meta ','
+      end
+      stra.meta i
     end
-    visitor.meta save
-    @keys.push save
   end
 
 end
@@ -781,29 +840,21 @@ class RecvH < StructH
   header_attr :for_,  String, R | W,        [:ford, :for_domain]
   header_attr :date,  Time,   R | W | NODUP
 
-
-  def eql?( oth )
-    super and
-    from  == oth.from and
-    by    == oth.by and
-    via   == oth.via and
-    with  == oth.with and
-    msgid == oth.msgid and
-    for_  == oth.for_ and
-    date  == oth.date
-  end
+  redefine_eql
+  empty_able
 
 
   private
 
   def init
-    super
+    @from = @by = @via = @with = @msgid = @for_ = nil
     @with = []
+    @date = nil
   end
 
   Tag = %w( from by via with id for )
 
-  def do_accept( visitor )
+  def do_accept( stra )
     val = [ @from, @by, @via ] + @with + [ @msgid, @for_ ]
     val.compact!
     c   = [ @from ? 1 : 0, @by ? 1 : 0, @via ? 1 : 0,
@@ -813,17 +864,18 @@ class RecvH < StructH
     c.each_with_index do |count, idx|
       label = Tag[idx]
       count.times do
-        visitor.spc unless i == 0
-        visitor.meta label
-        visitor.spc
-        visitor.meta val[i]
+        stra.spc unless i == 0
+        stra.meta label
+        stra.spc
+        stra.meta val[i]
         i += 1
       end
     end
+
+    stra.meta ';'
     if @date then
-      visitor.meta ';'
-      visitor.spc
-      visitor.meta ::TMail.time2str( @date )
+      stra.spc
+      stra.meta ::TMail.time2str( @date )
     end
   end
 
@@ -843,13 +895,18 @@ class MsgidH < StructH
 
   header_attr :msgid, String, R | W
   
-  def eql?( oth )
-    super and
-    msgid == oth.msgid
+  redefine_eql
+  empty_able
+
+
+  private
+
+  def init
+    @msgid = nil
   end
 
-  def do_accept( visitor )
-    visitor.meta @msgid
+  def do_accept( stra )
+    stra.meta @msgid
   end
 
 end
@@ -864,24 +921,24 @@ class EncH < StructH
   header_attr :encrypter, String, R | W
   header_attr :keyword,   String, R | W
 
-  def eql?( oth )
-    super and
-    self.encrypter == oth.encrypter and
-    self.keyword   == oth.keyword
-  end
+  redefine_eql
+  empty_able
+
+
+  private
 
   def init
-    super
+    @encrypter = nil
     @keyword = nil
   end
 
-  def do_accept( visitor )
+  def do_accept( stra )
     if @key then
-      visitor.meta @encrypter + ','
-      visitor.spc
-      visitor.meta @key
+      stra.meta @encrypter + ','
+      stra.spc
+      stra.meta @keyword
     else
-      visitor.meta @encrypter
+      stra.meta @encrypter
     end
   end
 
@@ -897,18 +954,23 @@ class VersionH < StructH
   header_attr :major, :Integer, R | W | NODUP
   header_attr :minor, :Integer, R | W | NODUP
 
+  redefine_eql
+  empty_able
+
   def version
     sprintf( '%d.%d', major, minor )
   end
 
-  def eql?( oth )
-    super and
-    self.major == oth.major and
-    self.minor == oth.minor
+  
+  private
+
+  def init
+    @major = nil
+    @minor = nil
   end
 
-  def do_accept( visitor )
-    visitor.meta "#{@major}.#{@minor}"
+  def do_accept( stra )
+    stra.meta "#{@major}.#{@minor}"
   end
 
 end
@@ -925,33 +987,29 @@ class CTypeH < StructH
   header_attr :sub,    :String, R | W
   header_attr :params, :Hash,   R | NODUP
 
+  redefine_eql
+  empty_able
+
   def []( key )
     params[key]
   end
 
-  def eql?( oth )
-    super and
-    main == oth.main and
-    sub  == oth.sub  and
-    params == oth.params
-  end
-
 
   private
 
   def init
-    super
+    @main = @sub = nil
     @params = {}
   end
 
-  def do_accept( visitor )
-    visitor.meta "#{@main}/#{@sub}"
+  def do_accept( stra )
+    stra.meta "#{@main}/#{@sub}"
     @params.each do |k,v|
-      visitor.meta ';'
-      visitor.spc
-      visitor.meta k
-      visitor.meta '='
-      visitor.text v
+      stra.meta ';'
+      stra.spc
+      stra.meta k
+      stra.meta '='
+      stra.value v
     end
   end
 
@@ -966,13 +1024,18 @@ class CEncodingH < StructH
 
   header_attr :encoding, :String, R | W
 
-  def eql?( oth )
-    super and
-    encoding == oth.encoding
+  redefine_eql
+  empty_able
+
+
+  private
+
+  def init
+    @encoding = nil
   end
 
-  def do_accept( visitor )
-    visitor.meta @encoding
+  def do_accept( stra )
+    stra.meta @encoding
   end
 
 end
@@ -991,28 +1054,25 @@ class CDispositionH < StructH
     params[key]
   end
 
-  def eql?( oth )
-    super and
-    self.disposition == oth.disposition and
-    self.params == oth.params
-  end
+  redefine_eql
+  empty_able
   
 
   private
 
   def init
-    super
+    @disposition = nil
     @params = {}
   end
 
-  def do_accept( visitor )
-    visitor.meta @disposition
+  def do_accept( stra )
+    stra.meta @disposition
     @params.each do |k,v|
-      visitor.meta ';'
-      visitor.spc
-      visitor.meta k
-      visitor.meta '='
-      visitor.text v
+      stra.meta ';'
+      stra.spc
+      stra.meta k
+      stra.meta '='
+      stra.value v
     end
   end
     
