@@ -23,7 +23,7 @@ end
 
 class FilePort < Port
 
-  FU = FileUtilities
+  FU = FILEUTILS
 
   def initialize( fname )
     @filename = File.expand_path( fname )
